"""Weaving code."""
import json
from pathlib import Path
from typing import Dict, Tuple
from multiprocessing import Pool

import tqdm
import numpy as np
from cv2 import cv2
from skimage.measure import profile_line

from pyanchorknit.utils import config_logger

logger = config_logger(__name__)

BLACK = 0
WHITE = 255


def whitest(intensities):
    return np.argmax(intensities)


def blackest(intensities):
    return np.argmin(intensities)


def create_circular_mask(h, w):
    center = (int(w / 2), int(h / 2))
    radius = min(center[0], center[1], w - center[0], h - center[1])

    Y, X = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)

    mask = dist_from_center <= radius
    return mask


def line_intensity(ps, points, bw_img):
    p1, p2 = ps
    return profile_line(bw_img.T, points[p1], points[p2], mode="constant").sum()


def weaving(
    img_path: Path,
    img_out: Path = Path("./weaved.png"),
    traces_out: Path = Path("./traces.json"),
    n_edges: int = 512,
    maxlines: int = 2000,
    n_jobs: int = 8,
):
    logger.info(f"image path: {img_path}")
    logger.info(f"number of edges: {n_edges}")
    logger.info(f"max lines: {maxlines}")
    logger.info(f"n_jobs: {n_jobs}")

    img = cv2.imread(str(img_path.resolve()))
    bw_img = 255 - cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    centerx = img.shape[0] // 2
    centery = img.shape[1] // 2

    points = []

    for theta in np.linspace(0, 2 * np.pi, n_edges + 1)[:-1]:
        x = min(int(centerx * np.cos(theta) + centerx), img.shape[0] - 1)
        y = min(int(centery * np.sin(theta) + centery), img.shape[1] - 1)
        points.append(np.array([x, y], dtype=np.int32))

    white_img: np.typing.ArrayLike = np.ones(img.shape[:2], dtype=np.uint8) * WHITE

    traces: Dict[Tuple[int, int], float] = dict()
    thread_length = 0

    p = 0
    max_intensity = None

    for _ in tqdm.tqdm(range(maxlines)):
        plines = [(p, j) for j in range(len(points)) if j != p]

        with Pool(n_jobs) as pool:
            intensities = pool.starmap(
                line_intensity, [(ps, points, bw_img) for ps in plines]
            )

        poi = whitest(intensities)
        p, pprime = plines[poi]
        intensity = intensities[poi]

        max_intensity = max(max_intensity or 0, intensity)  # flip by color

        if len(traces) > 0 and (intensity / max_intensity) < 0.10:
            logger.debug("Intensity below threshold")
            break

        if (p, pprime) in traces or (pprime, p) in traces:
            logger.debug("Repeated trace")
            break

        p1, p2 = tuple(points[p]), tuple(points[pprime])

        bw_img = cv2.line(bw_img, p1, p2, BLACK, 1)
        white_img = cv2.line(white_img, p1, p2, BLACK, 1)

        thread_length += np.linalg.norm(points[p] - points[pprime])
        traces[(p, pprime)] = intensity
        p = pprime

    logger.info(f"Thread used: {thread_length:.2f}")

    circular_mask = create_circular_mask(*img.shape[:2]) * 255

    rgba_img = cv2.merge(np.array([white_img, white_img, white_img, circular_mask]))
    cv2.imwrite(str(img_out.resolve()), rgba_img)

    points_json = {i: tuple(map(int, point)) for i, point in enumerate(points)}
    traces_json = [((x, y), trace_len) for (x, y), trace_len in traces.items()]
    with open(traces_out, "w") as fp:
        json.dump({"points": points_json, "traces_json": traces_json}, fp)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyanchorknit', 'pyanchorknit.cli']

package_data = \
{'': ['*']}

install_requires = \
['opencv-python>=4.5.3,<5.0.0',
 'scikit-image>=0.18.3,<0.19.0',
 'tqdm>=4.62.3,<5.0.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['pyanchorknit = pyanchorknit.cli.main:app()']}

setup_kwargs = {
    'name': 'pyanchorknit',
    'version': '0.1.0',
    'description': 'Weaving algorithm.',
    'long_description': '# PyAnchorKnit\n\n# Dev install\n## Poetry\n### Install pipx\n```bash\npython3 -m pip install --user pipx\npython3 -m pipx ensurepath\npipx completions\n```\n\n### Install poetry\n```bash\npipx install poetry\n```\n\n### Install the project\n```bash\npoetry install\n```\n\n## Install [pre-commit](https://pre-commit.com) hooks\n```bash\npoetry run pre-commit install\npoetry run pre-commit install -t pre-push\n```\n',
    'author': 'CrossNox',
    'author_email': 'ijmermet@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path

import typer

from pyanchorknit.weaving import weaving

app = typer.Typer()


@app.command()
def main(
    img: Path = typer.Argument(..., help="Path to input image"),
    img_out: Path = typer.Option("./weaved.png", help="Path to output image"),
    traces_out: Path = typer.Option(
        "./traces.json", help="Path to output json of traces"
    ),
    n_edges: int = typer.Option(512, help="Number of points around the edge"),
    maxlines: int = typer.Option(2000, help="Max number of traces"),
    n_jobs: int = typer.Option(8, help="Max number of jobs to use"),
):
    """Make weaving image."""
    weaving(
        img,
        img_out=img_out,
        traces_out=traces_out,
        n_edges=n_edges,
        maxlines=maxlines,
        n_jobs=n_jobs,
    )



"""Random utils."""
import logging

import typer


class TyperLoggerHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        fg = None
        bg = None
        if record.levelno == logging.DEBUG:
            fg = typer.colors.BLACK
            bg = typer.colors.WHITE
        elif record.levelno == logging.INFO:
            fg = typer.colors.BRIGHT_BLUE
        elif record.levelno == logging.WARNING:
            fg = typer.colors.BRIGHT_MAGENTA
        elif record.levelno == logging.CRITICAL:
            fg = typer.colors.BRIGHT_RED
        elif record.levelno == logging.ERROR:
            fg = typer.colors.BRIGHT_WHITE
            bg = typer.colors.BRIGHT_RED
        typer.secho(self.format(record), bg=bg, fg=fg)


def config_logger(name: str, level: int = logging.DEBUG, to_file: bool = False):
    """Configure logging for stream and file."""
    logger = logging.getLogger(name)
    logger.setLevel(level)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    # console_handler = logging.StreamHandler()
    # console_handler.setLevel(level)
    # console_handler.setFormatter(formatter)
    # logger.addHandler(console_handler)

    typer_handler = TyperLoggerHandler()
    typer_handler.setLevel(level)
    typer_handler.setFormatter(formatter)
    logger.addHandler(typer_handler)

    if to_file:
        try:
            file_handler = logging.FileHandler(f"{name}.log")
            file_handler.setLevel(logging.DEBUG)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
        except OSError:
            pass

    return logger
